from __future__ import annotations

import warnings
from typing import Iterable, Iterator, List, MutableSequence, Optional, overload

from attrs import define, field
from fontTools.pens.basePen import AbstractPen
from fontTools.pens.pointPen import AbstractPointPen, PointToSegmentPen

from ufoLib2.objects.misc import BoundingBox, getBounds, getControlBounds
from ufoLib2.objects.point import Point
from ufoLib2.serde import serde
from ufoLib2.typing import GlyphSet


@serde
@define
class Contour(MutableSequence[Point]):
    """Represents a contour as a list of points.

    Behavior:
        The Contour object has list-like behavior. This behavior allows you to interact
        with point data directly. For example, to get a particular point::

            point = contour[0]

        To iterate over all points::

            for point in contour:
                ...

        To get the number of points::

            pointCount = len(contour)

        To delete a particular point::

            del contour[0]

        To set a particular point to another Point object::

            contour[0] = anotherPoint
    """

    points: List[Point] = field(factory=list)
    """The list of points in the contour."""

    identifier: Optional[str] = field(default=None, repr=False)
    """The globally unique identifier of the contour."""

    # collections.abc.MutableSequence interface

    def __delitem__(self, index: int | slice) -> None:
        del self.points[index]

    @overload
    def __getitem__(self, index: int) -> Point: ...

    @overload
    def __getitem__(self, index: slice) -> list[Point]:  # noqa: F811
        ...

    def __getitem__(self, index: int | slice) -> Point | list[Point]:  # noqa: F811
        return self.points[index]

    def __setitem__(  # noqa: F811
        self, index: int | slice, point: Point | Iterable[Point]
    ) -> None:
        if isinstance(index, int) and isinstance(point, Point):
            self.points[index] = point
        elif (
            isinstance(index, slice)
            and isinstance(point, Iterable)
            and all(isinstance(p, Point) for p in point)
        ):
            self.points[index] = point
        else:
            raise TypeError(
                f"Expected Point or Iterable[Point], found {type(point).__name__}."
            )

    def __iter__(self) -> Iterator[Point]:
        return iter(self.points)

    def __len__(self) -> int:
        return len(self.points)

    def insert(self, index: int, value: Point) -> None:
        """Insert Point object ``value`` into the contour at ``index``."""
        if not isinstance(value, Point):
            raise TypeError(f"Expected Point, found {type(value).__name__}.")
        self.points.insert(index, value)

    # TODO: rotate method?

    @property
    def open(self) -> bool:
        """Returns whether the contour is open or closed."""
        if not self.points:
            return True
        return self.points[0].type == "move"

    def move(self, delta: tuple[float, float]) -> None:
        """Moves contour by (x, y) font units."""
        for point in self.points:
            point.move(delta)

    def getBounds(self, layer: GlyphSet | None = None) -> BoundingBox | None:
        """Returns the (xMin, yMin, xMax, yMax) bounding box of the glyph,
        taking the actual contours into account.

        Args:
            layer: Not applicable to contours, here for API symmetry.
        """
        return getBounds(self, layer)

    @property
    def bounds(self) -> BoundingBox | None:
        """Returns the (xMin, yMin, xMax, yMax) bounding box of the glyph,
        taking the actual contours into account.

        |defcon_compat|
        """
        return self.getBounds()

    def getControlBounds(self, layer: GlyphSet | None = None) -> BoundingBox | None:
        """Returns the (xMin, yMin, xMax, yMax) bounding box of the glyph,
        taking only the control points into account.

        Args:
            layer: Not applicable to contours, here for API symmetry.
        """
        return getControlBounds(self, layer)

    @property
    def controlPointBounds(self) -> BoundingBox | None:
        """Returns the (xMin, yMin, xMax, yMax) bounding box of the glyph,
        taking only the control points into account.

        |defcon_compat|
        """
        return self.getControlBounds()

    # -----------
    # Pen methods
    # -----------

    def draw(self, pen: AbstractPen) -> None:
        """Draws contour into given pen."""
        pointPen = PointToSegmentPen(pen)
        self.drawPoints(pointPen)

    def drawPoints(self, pointPen: AbstractPointPen) -> None:
        """Draws points of contour into given point pen."""
        try:
            pointPen.beginPath(identifier=self.identifier)
            for p in self.points:
                pointPen.addPoint(
                    (p.x, p.y),
                    segmentType=p.type,
                    smooth=p.smooth,
                    name=p.name,
                    identifier=p.identifier,
                )
        except TypeError:
            pointPen.beginPath()
            for p in self.points:
                pointPen.addPoint(
                    (p.x, p.y), segmentType=p.type, smooth=p.smooth, name=p.name
                )
            warnings.warn(
                "The pointPen needs an identifier kwarg. "
                "Identifiers have been discarded.",
                UserWarning,
            )
        pointPen.endPath()



from __future__ import annotations

from typing import BinaryIO, cast

from ufoLib2.typing import PathLike


def read_bytes(fp: PathLike | BinaryIO) -> bytes:
    if hasattr(fp, "read"):
        fp = cast(BinaryIO, fp)
        return fp.read()
    else:
        fp = cast(PathLike, fp)
        with open(fp, "rb") as f:
            return f.read()


def write_bytes(fp: PathLike | BinaryIO, data: bytes) -> None:
    if hasattr(fp, "write"):
        fp = cast(BinaryIO, fp)
        fp.write(data)
    else:
        fp = cast(PathLike, fp)
        with open(fp, "wb") as f:
            f.write(data)



from __future__ import annotations

from functools import partialmethod
from importlib import import_module
from typing import IO, Any, AnyStr, Callable, Type

from ufoLib2.errors import ExtrasNotInstalledError
from ufoLib2.typing import PathLike, T

_SERDE_FORMATS_ = ("json", "msgpack")


# the @serde decorator sets these as @classmethods on the class, hence
# the cls argument must appear as the first argument; in the standalone loads/load
# functions the input string or file-like object is the first argument and the second
# argument is the object_class, so below just swap the order of the arguments
def _loads(
    cls: Type[T], s: str | bytes, *, __callback: Callable[..., T], **kwargs: Any
) -> T:
    return __callback(s, cls, **kwargs)


def _load(
    cls: Type[T],
    fp: PathLike | IO[AnyStr],
    *,
    __callback: Callable[..., T],
    **kwargs: Any,
) -> T:
    return __callback(fp, cls, **kwargs)


def serde(cls: Type[T]) -> Type[T]:
    """Decorator to add serialization support to a ufoLib2 class.

    This adds f"{format}_loads" / f"{format}_dumps" (from/to bytes) methods, and
    f"{format}_load" / f"{format}_dump" (for file or path) methods to all ufoLib2
    objects, not just Font.

    Currently the supported formats are JSON and MessagePack (msgpack), but other
    formats may be added in the future.

    E.g.::

        from ufoLib2 import Font

        font = Font.open("MyFont.ufo")
        font.json_dump("MyFont.json")
        font2 = Font.json_load("MyFont.json")
        font3 = Font.json_loads(font2.json_dumps())

        font3.msgpack_dump("MyFont.msgpack")
        font4 = Font.msgpack_load("MyFont.msgpack")
        # etc.

    Note this requires additional extras to be installed: e.g. ufoLib2[json,msgpack].
    In additions to the respective serialization library, these installs the `cattrs`
    library for structuring/unstructuring custom objects from/to serializable data
    structures (also available separately as ufoLib2[converters] extra).

    If any of the optional dependencies fails to be imported, the methods will raise
    an ImportError when called.

    If the faster `orjson` library is present, it will be used in place of the
    built-in `json` library on CPython. On PyPy, the `orjson` library is not available,
    so the built-in `json` library will be used (though it's pretty fast anyway).

    If you want a serialization format that works out of the box with all ufoLib2
    objects (but it's mostly limited to Python) you can use the built-in pickle module,
    which doesn't require to use the `cattrs` converters.

    """

    supported_formats = []
    for fmt in _SERDE_FORMATS_:
        try:
            serde_submodule = import_module(f"ufoLib2.serde.{fmt}")
        except ImportError as exc:
            raise_error = ExtrasNotInstalledError(fmt)
            raise_error.__cause__ = exc
            for method in ("loads", "load", "dumps", "dump"):
                setattr(cls, f"{fmt}_{method}", raise_error)
        else:
            setattr(
                cls,
                f"{fmt}_loads",
                partialmethod(classmethod(_loads), __callback=serde_submodule.loads),  # type: ignore[arg-type]
            )
            setattr(
                cls,
                f"{fmt}_load",
                partialmethod(classmethod(_load), __callback=serde_submodule.load),  # type: ignore[arg-type]
            )
            setattr(cls, f"{fmt}_dumps", serde_submodule.dumps)
            setattr(cls, f"{fmt}_dump", serde_submodule.dump)
            supported_formats.append(fmt)

    setattr(cls, "_SERDE_FORMATS_", tuple(supported_formats))

    return cls



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    KeysView,
    Optional,
    Sequence,
    Type,
    overload,
)

from attrs import define, field
from fontTools.ufoLib.glifLib import GlyphSet

from ufoLib2.constants import DEFAULT_LAYER_NAME
from ufoLib2.objects.glyph import Glyph
from ufoLib2.objects.lib import (
    Lib,
    _convert_Lib,
    _get_lib,
    _get_tempLib,
    _set_lib,
    _set_tempLib,
)
from ufoLib2.objects.misc import (
    BoundingBox,
    _deepcopy_unlazify_attrs,
    _getstate_unlazify_attrs,
    _prune_object_libs,
    _setstate_attrs,
    unionBounds,
)
from ufoLib2.serde import serde
from ufoLib2.typing import T

if TYPE_CHECKING:
    from cattrs import Converter

_GLYPH_NOT_LOADED = Glyph(name="___UFOLIB2_LAZY_GLYPH___")


def _convert_glyphs(value: dict[str, Glyph] | Sequence[Glyph]) -> dict[str, Glyph]:
    result: dict[str, Glyph] = {}
    if isinstance(value, dict):
        glyph_ids = set()
        for name, glyph in value.items():
            if not isinstance(glyph, Glyph):
                raise TypeError(f"Expected Glyph, found {type(glyph).__name__}")
            if glyph is not _GLYPH_NOT_LOADED:
                glyph_id = id(glyph)
                if glyph_id in glyph_ids:
                    raise KeyError(f"{glyph!r} can't be added twice")
                glyph_ids.add(glyph_id)
                if glyph.name is None:
                    glyph._name = name
                elif glyph.name != name:
                    raise ValueError(
                        "glyph has incorrect name: "
                        f"expected '{name}', found '{glyph.name}'"
                    )
            result[name] = glyph
    else:
        for glyph in value:
            if not isinstance(glyph, Glyph):
                raise TypeError(f"Expected Glyph, found {type(glyph).__name__}")
            if glyph.name is None:
                raise ValueError(f"{glyph!r} has no name; can't add it to Layer")
            if glyph.name in result:
                raise KeyError(f"glyph named '{glyph.name}' already exists")
            result[glyph.name] = glyph
    return result


@serde
@define
class Layer:
    """Represents a Layer that holds Glyph objects.

    See http://unifiedfontobject.org/versions/ufo3/glyphs/layerinfo.plist/.

    Note:
        Various methods that work on Glyph objects take a ``layer`` attribute, because
        the UFO data model prescribes that Components within a Glyph object refer to
        glyphs *within the same layer*.

    Behavior:
        Layer behaves **partly** like a dictionary of type ``Dict[str, Glyph]``.
        Unless the font is loaded eagerly (with ``lazy=False``), the Glyph objects
        by default are only loaded into memory when accessed.

        To get the number of glyphs in the layer::

            glyphCount = len(layer)

        To iterate over all glyphs::

            for glyph in layer:
                ...

        To check if a specific glyph exists::

            exists = "myGlyphName" in layer

        To get a specific glyph::

            layer["myGlyphName"]

        To delete a specific glyph::

            del layer["myGlyphName"]
    """

    _name: str = field(default=DEFAULT_LAYER_NAME, metadata={"omit_if_default": False})
    _glyphs: Dict[str, Glyph] = field(factory=dict, converter=_convert_glyphs)
    color: Optional[str] = None
    """The color assigned to the layer."""

    _lib: Lib = field(factory=Lib, converter=_convert_Lib)
    """The layer's lib for mapping string keys to arbitrary data."""

    _default: bool = False
    """Can set to True to mark a layer as default. If layer name is 'public.default'
    the default attribute is automatically True. Exactly one layer must be marked as
    default in a font."""

    _tempLib: Lib = field(factory=Lib, converter=_convert_Lib)
    """A temporary map of arbitrary plist values."""

    _lazy: Optional[bool] = field(default=None, init=False, eq=False)
    _glyphSet: Any = field(default=None, init=False, eq=False)

    def __attrs_post_init__(self) -> None:
        if self._name == DEFAULT_LAYER_NAME and not self._default:
            # layer named 'public.default' is default by definition
            self._default = True

    @classmethod
    def read(
        cls, name: str, glyphSet: GlyphSet, lazy: bool = True, default: bool = False
    ) -> Layer:
        """Instantiates a Layer object from a
        :class:`fontTools.ufoLib.glifLib.GlyphSet`.

        Args:
            name: The name of the layer.
            glyphSet: The GlyphSet object to read from.
            lazy: If True, load glyphs as they are accessed. If False, load everything
                up front.
        """
        glyphNames = glyphSet.keys()
        glyphs: dict[str, Glyph]
        if lazy:
            glyphs = {gn: _GLYPH_NOT_LOADED for gn in glyphNames}
        else:
            glyphs = {}
            for glyphName in glyphNames:
                glyph = Glyph(glyphName)
                glyphSet.readGlyph(glyphName, glyph, glyph.getPointPen())
                glyphs[glyphName] = glyph
        self = cls(name, glyphs, default=default)
        self._lazy = lazy
        if lazy:
            self._glyphSet = glyphSet
        glyphSet.readLayerInfo(self)
        return self

    def unlazify(self) -> None:
        """Load all glyphs into memory."""
        if self._lazy:
            for _ in self:
                pass
        self._lazy = False

    __deepcopy__ = _deepcopy_unlazify_attrs

    __getstate__ = _getstate_unlazify_attrs
    __setstate__ = _setstate_attrs

    def __contains__(self, name: object) -> bool:
        return name in self._glyphs

    def __delitem__(self, name: str) -> None:
        del self._glyphs[name]

    def __getitem__(self, name: str) -> Glyph:
        glyph_object = self._glyphs[name]
        if glyph_object is _GLYPH_NOT_LOADED:
            return self.loadGlyph(name)
        return glyph_object

    def __setitem__(self, name: str, glyph: Glyph) -> None:
        if not isinstance(glyph, Glyph):
            raise TypeError(f"Expected Glyph, found {type(glyph).__name__}")
        glyph._name = name
        self._glyphs[name] = glyph

    def __iter__(self) -> Iterator[Glyph]:
        for name in self._glyphs:
            yield self[name]

    def __len__(self) -> int:
        return len(self._glyphs)

    def __repr__(self) -> str:
        n = len(self._glyphs)
        return "<{}.{} '{}' ({}{}) at {}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            self._name,
            "default, " if self._default else "",
            "empty" if n == 0 else "{} glyph{}".format(n, "s" if n > 1 else ""),
            hex(id(self)),
        )

    def get(self, name: str, default: T | None = None) -> T | Glyph | None:
        """Return the Glyph object for name if it is present in this layer,
        otherwise return ``default``."""
        try:
            return self[name]
        except KeyError:
            return default

    def keys(self) -> KeysView[str]:
        """Returns a list of glyph names."""
        return self._glyphs.keys()

    @overload
    def pop(self, key: str) -> Glyph: ...

    @overload
    def pop(self, key: str, default: Glyph | T = ...) -> Glyph | T: ...

    def pop(self, key: str, default: Glyph | T = KeyError) -> Glyph | T:  # type: ignore
        """Remove and return glyph from layer.

        Args:
            key: The name of the glyph.
            default: What to return if there is no glyph with the given name.
        """
        # NOTE: We can't defer to self._glyphs.pop because we must load glyphs
        try:
            glyph = self[key]
        except KeyError:
            if default is KeyError:
                raise
            glyph = default  # type: ignore
        else:
            del self[key]
        return glyph

    @property
    def name(self) -> str:
        """The name of the layer."""
        return self._name

    lib = property(_get_lib, _set_lib)

    tempLib = property(_get_tempLib, _set_tempLib)

    @property
    def default(self) -> bool:
        """Read-only property. To change the font's default layer use the
        LayerSet.defaultLayer property setter."""
        return self._default

    @property
    def bounds(self) -> BoundingBox | None:
        """Returns the (xMin, yMin, xMax, yMax) bounding box of the layer,
        taking the actual contours into account.

        |defcon_compat|
        """
        bounds = None
        for glyph in self:
            bounds = unionBounds(bounds, glyph.getBounds(self))
        return bounds

    @property
    def controlPointBounds(self) -> BoundingBox | None:
        """Returns the (xMin, yMin, xMax, yMax) bounding box of the layer,
        taking only the control points into account.

        |defcon_compat|
        """
        bounds = None
        for glyph in self:
            bounds = unionBounds(bounds, glyph.getControlBounds(self))
        return bounds

    def addGlyph(self, glyph: Glyph) -> None:
        """Appends glyph object to the this layer unless its name is already
        taken."""
        self.insertGlyph(glyph, overwrite=False, copy=False)

    def insertGlyph(
        self,
        glyph: Glyph,
        name: str | None = None,
        overwrite: bool = True,
        copy: bool = True,
    ) -> None:
        """Inserts Glyph object into this layer.

        Args:
            glyph: The Glyph object.
            name: The name of the glyph.
            overwrite: If True, overwrites (read: deletes) glyph with the same name if
                it exists. If False, raises KeyError.
            copy: If True, copies the Glyph object before insertion. If False, inserts
                as is.
        """
        if copy:
            glyph = glyph.copy()
        if name is not None:
            glyph._name = name
        if glyph.name is None:
            raise ValueError(f"{glyph!r} has no name; can't add it to Layer")
        if not overwrite and glyph.name in self._glyphs:
            raise KeyError(f"glyph named '{glyph.name}' already exists")
        self._glyphs[glyph.name] = glyph

    def loadGlyph(self, name: str) -> Glyph:
        """Load and return Glyph object."""
        # XXX: Remove and let __getitem__ do it?
        glyph = Glyph(name)
        self._glyphSet.readGlyph(name, glyph, glyph.getPointPen())
        self._glyphs[name] = glyph
        return glyph

    def newGlyph(self, name: str) -> Glyph:
        """Creates and returns new Glyph object in this layer with name."""
        if name in self._glyphs:
            raise KeyError(f"glyph named '{name}' already exists")
        self._glyphs[name] = glyph = Glyph(name)
        return glyph

    def renameGlyph(self, name: str, newName: str, overwrite: bool = False) -> None:
        """Renames a Glyph object in this layer.

        Args:
            name: The old name.
            newName: The new name.
            overwrite: If False, raises exception if newName is already taken.
                If True, overwrites (read: deletes) the old Glyph object.
        """
        if name == newName:
            return
        if not overwrite and newName in self._glyphs:
            raise KeyError(f"target glyph named '{newName}' already exists")
        # pop and set name
        glyph = self.pop(name)
        glyph._name = newName
        # add it back
        self._glyphs[newName] = glyph

    def instantiateGlyphObject(self) -> Glyph:
        """Returns a new Glyph instance.

        |defcon_compat|
        """
        return Glyph()

    def write(self, glyphSet: GlyphSet, saveAs: bool = True) -> None:
        """Write Layer to a :class:`fontTools.ufoLib.glifLib.GlyphSet`.

        Args:
            glyphSet: The GlyphSet object to write to.
            saveAs: If True, tells the writer to save out-of-place. If False, tells the
                writer to save in-place. This affects how resources are cleaned before
                writing.
        """
        glyphs = self._glyphs
        if not saveAs:
            for name in set(glyphSet.contents).difference(glyphs):
                glyphSet.deleteGlyph(name)
        for name, glyph in glyphs.items():
            if glyph is _GLYPH_NOT_LOADED:
                if saveAs:
                    glyph = self.loadGlyph(name)
                else:
                    continue
            _prune_object_libs(glyph.lib, _fetch_glyph_identifiers(glyph))
            glyphSet.writeGlyph(
                name, glyphObject=glyph, drawPointsFunc=glyph.drawPoints
            )
        glyphSet.writeContents()
        glyphSet.writeLayerInfo(self)
        if saveAs:
            # all glyphs are loaded by now, no need to keep ref to glyphSet
            self._glyphSet = None

    def _unstructure(self, converter: Converter) -> dict[str, Any]:
        # omit glyph name attribute, already used as key
        glyphs: dict[str, dict[str, Any]] = {}
        for glyph_name in self._glyphs:
            g = converter.unstructure(self[glyph_name])
            assert glyph_name == g.pop("name")
            glyphs[glyph_name] = g
        d: dict[str, Any] = {
            # never omit name even if == 'public.default' as that acts as
            # the layer's "key" in the layerSet.
            "name": self._name,
        }
        default: Any
        for key, value, default in [
            ("default", self._default, self._name == DEFAULT_LAYER_NAME),
            ("glyphs", glyphs, {}),
            ("lib", self._lib, {}),
            ("tempLib", self._tempLib, {}),
        ]:
            if not converter.omit_if_default or value != default:
                d[key] = value
        if self.color is not None:
            d["color"] = self.color
        return d

    @staticmethod
    def _structure(
        data: dict[str, Any], cls: Type[Layer], converter: Converter
    ) -> Layer:
        return cls(
            name=data.get("name", DEFAULT_LAYER_NAME),
            glyphs={
                k: converter.structure(v, Glyph)
                for k, v in data.get("glyphs", {}).items()
            },
            color=data.get("color"),
            lib=converter.structure(data.get("lib", {}), Lib),
            default=data.get("default", False),
            tempLib=converter.structure(data.get("tempLib", {}), Lib),
        )


def _fetch_glyph_identifiers(glyph: Glyph) -> set[str]:
    """Returns all identifiers in use in a glyph."""

    identifiers = set()
    for anchor in glyph.anchors:
        if anchor.identifier is not None:
            identifiers.add(anchor.identifier)
    for guideline in glyph.guidelines:
        if guideline.identifier is not None:
            identifiers.add(guideline.identifier)
    for contour in glyph.contours:
        if contour.identifier is not None:
            identifiers.add(contour.identifier)
        for point in contour:
            if point.identifier is not None:
                identifiers.add(point.identifier)
    for component in glyph.components:
        if component.identifier is not None:
            identifiers.add(component.identifier)
    return identifiers



"""Fontinfo.plist fields for WOFF 1.0 metadata.

https://unifiedfontobject.org/versions/ufo3/fontinfo.plist/#woff-data
"""

from __future__ import annotations

from typing import Any, List, Mapping, Optional, Sequence, Type, TypeVar

from attrs import Attribute, define, field

from ufoLib2.objects.misc import AttrDictMixin

_T = TypeVar("_T", bound=AttrDictMixin)


def _convert_list_of_woff_metadata(
    cls: Type[_T], values: Sequence[_T | Mapping[str, Any]]
) -> list[_T]:
    return [cls.coerce_from_dict(v) for v in values]


@define
class WoffMetadataUniqueID(AttrDictMixin):
    id: str


@define
class WoffMetadataVendor(AttrDictMixin):
    name: str
    url: Optional[str] = None
    dir: Optional[str] = None
    # 'class' of course is reserved in Python
    class_: Optional[str] = field(default=None, metadata={"rename_attr": "class"})


@define
class WoffMetadataCredit(AttrDictMixin):
    name: str
    url: Optional[str] = None
    role: Optional[str] = None
    dir: Optional[str] = None
    class_: Optional[str] = field(default=None, metadata={"rename_attr": "class"})


def _convert_list_of_woff_metadata_credits(
    value: list[WoffMetadataCredit | Mapping[str, Any]],
) -> list[WoffMetadataCredit]:
    return _convert_list_of_woff_metadata(WoffMetadataCredit, value)


@define
class WoffMetadataCredits(AttrDictMixin):
    credits: List[WoffMetadataCredit] = field(
        factory=list,
        converter=_convert_list_of_woff_metadata_credits,
    )


@define
class WoffMetadataText(AttrDictMixin):
    text: str
    language: Optional[str] = None
    dir: Optional[str] = None
    class_: Optional[str] = field(default=None, metadata={"rename_attr": "class"})


def _at_least_one_item(
    self: Any, attribute: Attribute[Any], value: Sequence[Any]
) -> None:
    if len(value) == 0:
        raise ValueError(
            f"{self.__class__.__name__}.{attribute.name} must contain at list 1 item"
        )


def _convert_list_of_woff_metadata_texts(
    value: list[WoffMetadataText | Mapping[str, Any]],
) -> list[WoffMetadataText]:
    return _convert_list_of_woff_metadata(WoffMetadataText, value)


@define
class WoffMetadataDescription(AttrDictMixin):
    url: Optional[str] = None
    text: List[WoffMetadataText] = field(
        factory=list,
        validator=_at_least_one_item,
        converter=_convert_list_of_woff_metadata_texts,
    )


@define
class WoffMetadataLicense(AttrDictMixin):
    url: Optional[str] = None
    id: Optional[str] = None
    text: List[WoffMetadataText] = field(
        factory=list,
        converter=_convert_list_of_woff_metadata_texts,
    )


@define
class WoffMetadataCopyright(AttrDictMixin):
    text: List[WoffMetadataText] = field(
        factory=list,
        validator=_at_least_one_item,
        converter=_convert_list_of_woff_metadata_texts,
    )


@define
class WoffMetadataTrademark(AttrDictMixin):
    text: List[WoffMetadataText] = field(
        factory=list,
        validator=_at_least_one_item,
        converter=_convert_list_of_woff_metadata_texts,
    )


@define
class WoffMetadataLicensee(AttrDictMixin):
    name: str
    dir: Optional[str] = None
    class_: Optional[str] = field(default=None, metadata={"rename_attr": "class"})


@define
class WoffMetadataExtensionName(AttrDictMixin):
    text: str
    language: Optional[str] = None
    dir: Optional[str] = None
    class_: Optional[str] = field(default=None, metadata={"rename_attr": "class"})


@define
class WoffMetadataExtensionValue(AttrDictMixin):
    text: str
    language: Optional[str] = None
    dir: Optional[str] = None
    class_: Optional[str] = field(default=None, metadata={"rename_attr": "class"})


def _convert_list_of_woff_metadata_extension_name(
    value: list[WoffMetadataExtensionName | Mapping[str, Any]],
) -> list[WoffMetadataExtensionName]:
    return _convert_list_of_woff_metadata(WoffMetadataExtensionName, value)


def _convert_list_of_woff_metadata_extension_value(
    value: list[WoffMetadataExtensionValue | Mapping[str, Any]],
) -> list[WoffMetadataExtensionValue]:
    return _convert_list_of_woff_metadata(WoffMetadataExtensionValue, value)


@define
class WoffMetadataExtensionItem(AttrDictMixin):
    id: Optional[str] = None
    names: List[WoffMetadataExtensionName] = field(
        factory=list,
        validator=_at_least_one_item,
        converter=_convert_list_of_woff_metadata_extension_name,
    )
    # 'values()' is the name of the dict method, hence the attribute named 'values_'
    values_: List[WoffMetadataExtensionValue] = field(
        factory=list,
        validator=_at_least_one_item,
        converter=_convert_list_of_woff_metadata_extension_value,
        metadata={"rename_attr": "values"},
    )


def _convert_list_of_woff_metadata_extension_item(
    value: list[WoffMetadataExtensionItem | Mapping[str, Any]],
) -> list[WoffMetadataExtensionItem]:
    return _convert_list_of_woff_metadata(WoffMetadataExtensionItem, value)


@define
class WoffMetadataExtension(AttrDictMixin):
    id: Optional[str]
    names: List[WoffMetadataExtensionName] = field(
        factory=list,
        converter=_convert_list_of_woff_metadata_extension_name,
    )
    # 'items()' is the name of the dict method, hence the attribute named 'items_'
    items_: List[WoffMetadataExtensionItem] = field(
        factory=list,
        validator=_at_least_one_item,
        converter=_convert_list_of_woff_metadata_extension_item,
        metadata={"rename_attr": "items"},
    )



from __future__ import annotations

from typing import Optional

from attrs import define

from ufoLib2.objects.misc import AttrDictMixin
from ufoLib2.serde import serde


@serde
@define
class Anchor(AttrDictMixin):
    """Represents a single anchor.

    See http://unifiedfontobject.org/versions/ufo3/glyphs/glif/#anchor.
    """

    x: float
    """The x coordinate of the anchor."""

    y: float
    """The y coordinate of the anchor."""

    name: Optional[str] = None
    """The name of the anchor."""

    color: Optional[str] = None
    """The color of the anchor."""

    identifier: Optional[str] = None
    """The globally unique identifier of the anchor."""

    def move(self, delta: tuple[float, float]) -> None:
        """Moves anchor by (x, y) font units."""
        x, y = delta
        self.x += x
        self.y += y



from __future__ import annotations

from typing import Optional

from attrs import define

from ufoLib2.serde import serde


@serde
@define
class Point:
    """Represents a single point.

    See http://unifiedfontobject.org/versions/ufo3/glyphs/glif/#point.
    """

    x: float
    """The x coordinate of the point."""

    y: float
    """The y coordinate of the point."""

    type: Optional[str] = None
    """The type of the point.

    ``None`` means "offcurve".

    See http://unifiedfontobject.org/versions/ufo3/glyphs/glif/#point-types.
    """

    smooth: bool = False
    """Whether a smooth curvature should be maintained at this point."""

    name: Optional[str] = None
    """The name of the point, no uniqueness required."""

    identifier: Optional[str] = None
    """The globally unique identifier of the point."""

    # XXX: Add post_init to check spec-mandated invariants?

    @property
    def segmentType(self) -> str | None:
        """Returns the type of the point.

        |defcon_compat|
        """
        return self.type

    def move(self, delta: tuple[float, float]) -> None:
        """Moves point by (x, y) font units."""
        x, y = delta
        self.x += x
        self.y += y



"""ufoLib2 -- a package for dealing with UFO fonts."""

from __future__ import annotations

from ufoLib2.objects import Font

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "0.0.0+unknown"


__all__ = ["Font"]



from __future__ import annotations

from copy import deepcopy
from typing import Any, Iterator, List, Mapping, Optional, cast

from attrs import define, field
from fontTools.misc.transform import Transform
from fontTools.pens.basePen import AbstractPen
from fontTools.pens.pointPen import (
    AbstractPointPen,
    PointToSegmentPen,
    SegmentToPointPen,
)

from ufoLib2.objects.anchor import Anchor
from ufoLib2.objects.component import Component
from ufoLib2.objects.contour import Contour
from ufoLib2.objects.guideline import Guideline
from ufoLib2.objects.image import Image
from ufoLib2.objects.lib import (
    Lib,
    _convert_Lib,
    _get_lib,
    _get_tempLib,
    _set_lib,
    _set_tempLib,
)
from ufoLib2.objects.misc import BoundingBox, _object_lib, getBounds, getControlBounds
from ufoLib2.pointPens.glyphPointPen import GlyphPointPen
from ufoLib2.serde import serde
from ufoLib2.typing import GlyphSet, HasIdentifier


@serde
@define
class Glyph:
    """Represents a glyph, containing contours, components, anchors and various
    other bits of data concerning it.

    See http://unifiedfontobject.org/versions/ufo3/glyphs/glif/.

    Behavior:
        The Glyph object has list-like behavior. This behavior allows you to interact
        with contour data directly. For example, to get a particular contour::

            contour = glyph[0]

        To iterate over all contours::

            for contour in glyph:
                ...

        To get the number of contours::

            contourCount = len(glyph)

        To check if a :class:`.Contour` object is in glyph::

            exists = contour in glyph

        To interact with components or anchors in a similar way, use the
        :attr:`.Glyph.components` and :attr:`.Glyph.anchors` attributes.
    """

    _name: Optional[str] = None

    width: float = 0
    """The width of the glyph."""

    height: float = 0
    """The height of the glyph."""

    unicodes: List[int] = field(factory=list)
    """The Unicode code points assigned to the glyph. Note that a glyph can have
    multiple."""

    _image: Image = field(factory=Image)

    _lib: Lib = field(factory=Lib, converter=_convert_Lib)
    """The glyph's mapping of string keys to arbitrary data."""

    note: Optional[str] = None
    """A free form text note about the glyph."""

    _anchors: List[Anchor] = field(factory=list)
    components: List[Component] = field(factory=list)
    """The list of components the glyph contains."""

    contours: List[Contour] = field(factory=list)
    """The list of contours the glyph contains."""

    _guidelines: List[Guideline] = field(factory=list)

    _tempLib: Lib = field(factory=Lib, converter=_convert_Lib)
    """A temporary map of arbitrary plist values."""

    def __len__(self) -> int:
        return len(self.contours)

    def __getitem__(self, index: int) -> Contour:
        return self.contours[index]

    def __contains__(self, contour: Contour) -> bool:
        return contour in self.contours

    def __iter__(self) -> Iterator[Contour]:
        return iter(self.contours)

    def __repr__(self) -> str:
        return "<{}.{} {}at {}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            f"'{self._name}' " if self._name is not None else "",
            hex(id(self)),
        )

    lib = property(_get_lib, _set_lib)

    tempLib = property(_get_tempLib, _set_tempLib)

    @property
    def anchors(self) -> list[Anchor]:
        """The list of anchors the glyph contains.

        Getter:
            Returns a list of anchors the glyph contains. Modifications of the list
            modify the Glyph object.

        Setter:
            Clears current anchors and sets the new ones.
        """
        return self._anchors

    @anchors.setter
    def anchors(self, value: list[Anchor]) -> None:
        self.clearAnchors()
        for anchor in value:
            self.appendAnchor(anchor)

    @property
    def guidelines(self) -> list[Guideline]:
        """The list of guidelines the glyph contains.

        Getter:
            Returns a list of guidelines the glyph contains. Modifications of the list
            modify the Glyph object.

        Setter:
            Clears current guidelines and sets the new ones.
        """
        return self._guidelines

    @guidelines.setter
    def guidelines(self, value: list[Guideline]) -> None:
        self.clearGuidelines()
        for guideline in value:
            self.appendGuideline(guideline)

    @property
    def name(self) -> str | None:
        """The name of the glyph."""
        return self._name

    @property
    def unicode(self) -> int | None:
        """The first assigned Unicode code point or None.

        See http://unifiedfontobject.org/versions/ufo3/glyphs/glif/#unicode.

        Setter:
            Sets the value to be the first of the assigned Unicode code points. Will
            remove a duplicate if exists. Will clear the list of Unicode points if
            value is None.
        """
        if self.unicodes:
            return self.unicodes[0]
        return None

    @unicode.setter
    def unicode(self, value: int | None) -> None:
        if value is None:
            self.unicodes = []
        else:
            if self.unicodes:
                if self.unicodes[0] == value:
                    return
                try:
                    self.unicodes.remove(value)
                except ValueError:
                    pass
                self.unicodes.insert(0, value)
            else:
                self.unicodes.append(value)

    @property
    def image(self) -> Image:
        """The background image reference associated with the glyph.

        See http://unifiedfontobject.org/versions/ufo3/glyphs/glif/#image.

        Setter:
            Sets the background image reference. Clears it if value is None.
        """
        return self._image

    @image.setter
    def image(self, image: Image | Mapping[str, Any] | None) -> None:
        if image is None:
            self._image.clear()
        elif isinstance(image, Image):
            self._image = image
        else:
            self._image = Image(
                fileName=image["fileName"],
                transformation=Transform(
                    image["xScale"],
                    image["xyScale"],
                    image["yxScale"],
                    image["yScale"],
                    image["xOffset"],
                    image["yOffset"],
                ),
                color=image.get("color"),
            )

    def objectLib(self, object: HasIdentifier) -> dict[str, Any]:
        """Return the lib for an object with an identifier, as stored in a glyph's lib.

        If the object does not yet have an identifier, a new one is assigned to it. If
        the font lib does not yet contain the object's lib, a new one is inserted and
        returned.

        .. doctest::

            >>> from ufoLib2.objects import Font, Guideline
            >>> font = Font()
            >>> glyph = font.newGlyph("a")
            >>> glyph.guidelines = [Guideline(x=100)]
            >>> guideline_lib = glyph.objectLib(glyph.guidelines[0])
            >>> guideline_lib["com.test.foo"] = 1234
            >>> guideline_id = glyph.guidelines[0].identifier
            >>> assert guideline_id is not None
            >>> assert glyph.lib["public.objectLibs"][guideline_id] is guideline_lib
        """

        return _object_lib(self.lib, object)

    def clear(self) -> None:
        """Clears out anchors, components, contours, guidelines and image
        references."""
        del self._anchors[:]
        del self.components[:]
        del self.contours[:]
        del self._guidelines[:]
        self.image.clear()

    def clearAnchors(self) -> None:
        """Clears out anchors."""
        del self._anchors[:]

    def clearContours(self) -> None:
        """Clears out contours."""
        del self.contours[:]

    def clearComponents(self) -> None:
        """Clears out components."""
        del self.components[:]

    def clearGuidelines(self) -> None:
        """Clears out guidelines."""
        del self._guidelines[:]

    def removeComponent(self, component: Component) -> None:
        """Removes :class:`.Component` object from the glyph's list of components."""
        self.components.remove(component)

    def appendAnchor(self, anchor: Anchor | Mapping[str, Any]) -> None:
        """Appends an :class:`.Anchor` object to glyph's list of anchors.

        Args:
            anchor: An :class:`.Anchor` object or mapping for the Anchor constructor.
        """
        if not isinstance(anchor, Anchor):
            if not isinstance(anchor, Mapping):
                raise TypeError(
                    "Expected Anchor object or a Mapping for the ",
                    f"Anchor constructor, found {type(anchor).__name__}",
                )
            anchor = Anchor(**anchor)
        self.anchors.append(anchor)

    def appendGuideline(self, guideline: Guideline | Mapping[str, Any]) -> None:
        """Appends a :class:`.Guideline` object to glyph's list of guidelines.

        Args:
            guideline: A :class:`.Guideline` object or a mapping for the Guideline
                constructor.
        """
        if not isinstance(guideline, Guideline):
            if not isinstance(guideline, Mapping):
                raise TypeError(
                    "Expected Guideline object or a Mapping for the ",
                    f"Guideline constructor, found {type(guideline).__name__}",
                )
            guideline = Guideline(**guideline)
        self._guidelines.append(guideline)

    def appendContour(self, contour: Contour) -> None:
        """Appends a :class:`.Contour` object to glyph's list of contours."""
        if not isinstance(contour, Contour):
            raise TypeError(f"Expected Contour, found {type(contour).__name__}")
        self.contours.append(contour)

    def copy(self, name: str | None = None) -> Glyph:
        """Returns a new Glyph (deep) copy, optionally override the new glyph
        name."""
        other = deepcopy(self)
        if name is not None:
            other._name = name
        return other

    def copyDataFromGlyph(self, glyph: Glyph) -> None:
        """Deep-copies everything from the other glyph into self, except for
        the name.

        Existing glyph data is overwritten.

        |defcon_compat|
        """
        self.width = glyph.width
        self.height = glyph.height
        self.unicodes = list(glyph.unicodes)
        self.image = deepcopy(glyph.image)
        self.note = glyph.note
        self.lib = deepcopy(glyph.lib)
        self.anchors = deepcopy(glyph.anchors)
        self.guidelines = deepcopy(glyph.guidelines)
        # NOTE: defcon's copyDataFromGlyph appends instead of overwrites here,
        # but we do the right thing, for consistency with the rest.
        self.clearContours()
        self.clearComponents()
        pointPen = self.getPointPen()
        glyph.drawPoints(pointPen)

    def move(self, delta: tuple[float, float]) -> None:
        """Moves all contours, components and anchors by (x, y) font units."""
        for contour in self.contours:
            contour.move(delta)
        for component in self.components:
            component.move(delta)
        for anchor in self.anchors:
            anchor.move(delta)

    # -----------
    # Pen methods
    # -----------

    def draw(self, pen: AbstractPen, outputImpliedClosingLine: bool = False) -> None:
        """Draws glyph into given pen."""
        # TODO: Document pen interface more or link to somewhere.
        pointPen = PointToSegmentPen(
            pen, outputImpliedClosingLine=outputImpliedClosingLine
        )
        self.drawPoints(pointPen)

    def drawPoints(self, pointPen: AbstractPointPen) -> None:
        """Draws points of glyph into given point pen."""
        for contour in self.contours:
            contour.drawPoints(pointPen)
        for component in self.components:
            component.drawPoints(pointPen)

    def getPen(self) -> AbstractPen:
        """Returns a pen for others to draw into self."""
        pen = SegmentToPointPen(self.getPointPen())
        return pen

    def getPointPen(self) -> AbstractPointPen:
        """Returns a point pen for others to draw points into self."""
        pointPen = GlyphPointPen(self)
        return pointPen

    # lib wrapped attributes

    @property
    def markColor(self) -> str | None:
        """The color assigned to the glyph.

        See http://unifiedfontobject.org/versions/ufo3/glyphs/glif/#publicmarkcolor.

        Getter:
            Returns the mark color or None.

        Setter:
            Sets the mark color. If value is None, deletes the key from the lib if
            present.
        """
        return cast(Optional[str], self.lib.get("public.markColor"))

    @markColor.setter
    def markColor(self, value: str | None) -> None:
        if value is not None:
            self.lib["public.markColor"] = value
        elif "public.markColor" in self.lib:
            del self.lib["public.markColor"]

    @property
    def verticalOrigin(self) -> float | None:
        """The vertical origin of the glyph.

        See http://unifiedfontobject.org/versions/ufo3/glyphs/glif/#publicverticalorigin.

        Getter:
            Returns the vertical origin or None.

        Setter:
            Sets the vertical origin. If value is None, deletes the key from the lib if
            present.
        """
        return cast(Optional[float], self.lib.get("public.verticalOrigin"))

    @verticalOrigin.setter
    def verticalOrigin(self, value: float | None) -> None:
        if value is not None:
            self.lib["public.verticalOrigin"] = value
        elif "public.verticalOrigin" in self.lib:
            del self.lib["public.verticalOrigin"]

    # bounds and side-bearings

    def getBounds(self, layer: GlyphSet | None = None) -> BoundingBox | None:
        """Returns the (xMin, yMin, xMax, yMax) bounding box of the glyph,
        taking the actual contours into account.

        Args:
            layer: The layer of the glyph to look up components, if any. Not needed for
                pure-contour glyphs.
        """
        if layer is None and self.components:
            raise TypeError("layer is required to compute bounds of components")

        return getBounds(self, layer)

    def getControlBounds(self, layer: GlyphSet | None = None) -> BoundingBox | None:
        """Returns the (xMin, yMin, xMax, yMax) bounding box of the glyph,
        taking only the control points into account.

        Args:
            layer: The layer of the glyph to look up components, if any. Not needed for
                pure-contour glyphs.
        """
        if layer is None and self.components:
            raise TypeError("layer is required to compute bounds of components")

        return getControlBounds(self, layer)

    def getLeftMargin(self, layer: GlyphSet | None = None) -> float | None:
        """Returns the the space in font units from the point of origin to the
        left side of the glyph.

        Args:
            layer: The layer of the glyph to look up components, if any. Not needed for
                pure-contour glyphs.
        """
        bounds = self.getBounds(layer)
        if bounds is None:
            return None
        return bounds.xMin

    def setLeftMargin(self, value: float, layer: GlyphSet | None = None) -> None:
        """Sets the the space in font units from the point of origin to the
        left side of the glyph.

        Args:
            value: The desired left margin in font units.
            layer: The layer of the glyph to look up components, if any. Not needed for
                pure-contour glyphs.
        """
        bounds = self.getBounds(layer)
        if bounds is None:
            return None
        diff = value - bounds.xMin
        if diff:
            self.width += diff
            self.move((diff, 0))

    def getRightMargin(self, layer: GlyphSet | None = None) -> float | None:
        """Returns the the space in font units from the glyph's advance width
        to the right side of the glyph.

        Args:
            layer: The layer of the glyph to look up components, if any. Not needed for
                pure-contour glyphs.
        """
        bounds = self.getBounds(layer)
        if bounds is None:
            return None
        return self.width - bounds.xMax

    def setRightMargin(self, value: float, layer: GlyphSet | None = None) -> None:
        """Sets the the space in font units from the glyph's advance width to
        the right side of the glyph.

        Args:
            value: The desired right margin in font units.
            layer: The layer of the glyph to look up components, if any. Not needed for
                pure-contour glyphs.
        """
        bounds = self.getBounds(layer)
        if bounds is None:
            return None
        self.width = bounds.xMax + value

    def getBottomMargin(self, layer: GlyphSet | None = None) -> float | None:
        """Returns the the space in font units from the bottom of the canvas to
        the bottom of the glyph.

        Args:
            layer: The layer of the glyph to look up components, if any. Not needed for
                pure-contour glyphs.
        """
        bounds = self.getBounds(layer)
        if bounds is None:
            return None
        if self.verticalOrigin is None:
            return bounds.yMin
        else:
            return bounds.yMin - (self.verticalOrigin - self.height)

    def setBottomMargin(self, value: float, layer: GlyphSet | None = None) -> None:
        """Sets the the space in font units from the bottom of the canvas to
        the bottom of the glyph.

        Args:
            value: The desired bottom margin in font units.
            layer: The layer of the glyph to look up components, if any. Not needed for
                pure-contour glyphs.
        """
        bounds = self.getBounds(layer)
        if bounds is None:
            return None
        # blindly copied from defcon Glyph._set_bottomMargin; not sure it's correct
        if self.verticalOrigin is None:
            oldValue = bounds.yMin
            self.verticalOrigin = self.height
        else:
            oldValue = bounds.yMin - (self.verticalOrigin - self.height)
        diff = value - oldValue
        if diff:
            self.height += diff

    def getTopMargin(self, layer: GlyphSet | None = None) -> float | None:
        """Returns the the space in font units from the top of the canvas to
        the top of the glyph.

        Args:
            layer: The layer of the glyph to look up components, if any. Not needed for
                pure-contour glyphs.
        """
        bounds = self.getBounds(layer)
        if bounds is None:
            return None
        if self.verticalOrigin is None:
            return self.height - bounds.yMax
        else:
            return self.verticalOrigin - bounds.yMax

    def setTopMargin(self, value: float, layer: GlyphSet | None = None) -> None:
        """Sets the the space in font units from the top of the canvas to the
        top of the glyph.

        Args:
            value: The desired top margin in font units.
            layer: The layer of the glyph to look up components, if any. Not needed for
                pure-contour glyphs.
        """
        bounds = self.getBounds(layer)
        if bounds is None:
            return
        if self.verticalOrigin is None:
            oldValue = self.height - bounds.yMax
        else:
            oldValue = self.verticalOrigin - bounds.yMax
        diff = value - oldValue
        if oldValue != value:
            # Is this still correct when verticalOrigin was not previously set?
            self.verticalOrigin = bounds.yMax + value
            self.height += diff



from __future__ import annotations

import json
from typing import Any, BinaryIO, Type

from ufoLib2.converters import structure, unstructure
from ufoLib2.serde.util import read_bytes, write_bytes
from ufoLib2.typing import PathLike, T

have_orjson = False
try:
    import orjson

    have_orjson = True
except ImportError:
    pass


def dumps(
    obj: Any,
    indent: int | None = None,
    sort_keys: bool = False,
    **kwargs: Any,
) -> bytes:
    data = unstructure(obj)

    if have_orjson:
        if indent is not None:
            if indent != 2:
                raise ValueError("indent must be 2 or None for orjson")
            kwargs["option"] = kwargs.pop("option", 0) | orjson.OPT_INDENT_2
        if sort_keys:
            kwargs["option"] = kwargs.pop("option", 0) | orjson.OPT_SORT_KEYS
        # orjson.dumps always returns bytes
        result = orjson.dumps(data, **kwargs)
    else:
        # built-in json.dumps returns a string, not bytes, hence the encoding
        s = json.dumps(data, indent=indent, sort_keys=sort_keys, **kwargs)
        result = s.encode("utf-8")
    return result


def loads(s: str | bytes, object_class: Type[T], **kwargs: Any) -> T:
    if have_orjson:
        data = orjson.loads(s, **kwargs)
    else:
        data = json.loads(s, **kwargs)
    return structure(data, object_class)


def dump(
    obj: Any,
    fp: PathLike | BinaryIO,
    indent: int | None = None,
    sort_keys: bool = False,
    **kwargs: Any,
) -> None:
    write_bytes(fp, dumps(obj, indent=indent, sort_keys=sort_keys, **kwargs))


def load(fp: PathLike | BinaryIO, object_class: Type[T], **kwargs: Any) -> T:
    return loads(read_bytes(fp), object_class, **kwargs)



from __future__ import annotations

from typing import Any, BinaryIO, Type, cast

import msgpack  # type: ignore

from ufoLib2.converters import binary_converter
from ufoLib2.serde.util import read_bytes, write_bytes
from ufoLib2.typing import PathLike, T


def dumps(obj: Any, **kwargs: Any) -> bytes:
    data = binary_converter.unstructure(obj)
    result = msgpack.packb(data, **kwargs)
    return cast(bytes, result)


def loads(s: bytes, object_class: Type[T], **kwargs: Any) -> T:
    data = msgpack.unpackb(s, **kwargs)
    return binary_converter.structure(data, object_class)


def dump(obj: Any, fp: PathLike | BinaryIO, **kwargs: Any) -> None:
    write_bytes(fp, dumps(obj, **kwargs))


def load(fp: PathLike | BinaryIO, object_class: Type[T], **kwargs: Any) -> T:
    return loads(read_bytes(fp), object_class, **kwargs)



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Dict,
    Iterable,
    Iterator,
    Optional,
    Sized,
)

from attrs import define, field
from fontTools.ufoLib import UFOReader, UFOWriter

from ufoLib2.constants import DEFAULT_LAYER_NAME
from ufoLib2.errors import Error
from ufoLib2.objects.layer import Layer
from ufoLib2.objects.misc import (
    _deepcopy_unlazify_attrs,
    _getstate_unlazify_attrs,
    _setstate_attrs,
)
from ufoLib2.serde import serde
from ufoLib2.typing import T

if TYPE_CHECKING:
    from typing import Type

    from cattrs import Converter

_LAYER_NOT_LOADED = Layer(name="___UFOLIB2_LAZY_LAYER___")


def _must_have_at_least_one_item(self: Any, attribute: Any, value: Sized) -> None:
    if not len(value):
        raise ValueError("value must have at least one item.")


@serde
@define
class LayerSet:
    """Represents a mapping of layer names to Layer objects.

    See http://unifiedfontobject.org/versions/ufo3/layercontents.plist/ for layer
    semantics.

    Behavior:
        LayerSet behaves **partly** like a dictionary of type ``Dict[str, Layer]``,
        but creating and loading layers is done through their own methods. Unless the
        font is loaded eagerly (with ``lazy=False``), the layer objects and their
        glyphs are by default only loaded into memory when accessed.

        To get the number of layers in the font::

            layerCount = len(font.layers)

        To iterate over all layers::

            for layer in font.layers:
                ...

        To check if a specific layer exists::

            exists = "myLayerName" in font.layers

        To get a specific layer::

            font.layers["myLayerName"]

        To delete a specific layer::

            del font.layers["myLayerName"]
    """

    _layers: Dict[str, Layer] = field(
        validator=_must_have_at_least_one_item,
    )

    _defaultLayer: Layer = field(default=_LAYER_NOT_LOADED, eq=False)

    _lazy: Optional[bool] = field(default=None, init=False, eq=False)
    _reader: Optional[UFOReader] = field(default=None, init=False, eq=False)

    def __attrs_post_init__(self) -> None:
        if self._defaultLayer == _LAYER_NOT_LOADED:
            found = False
            for layer in self._layers.values():
                if layer._default:
                    if found:
                        raise ValueError("more than one layer marked as default")
                    found = True
                    self._defaultLayer = layer
            if not found:
                raise ValueError("no layer marked as default")
        else:
            if not any(layer is self._defaultLayer for layer in self._layers.values()):
                raise ValueError(
                    f"default layer {repr(self._defaultLayer)} must be in layer set."
                )
            assert self._defaultLayer._default

    @classmethod
    def default(cls) -> LayerSet:
        """Return a new LayerSet with an empty default Layer."""
        return cls.from_iterable([Layer()])

    @classmethod
    def from_iterable(
        cls, value: Iterable[Layer], defaultLayerName: str = DEFAULT_LAYER_NAME
    ) -> LayerSet:
        """Instantiates a LayerSet from an iterable of :class:`.Layer` objects.

        Args:
            value: an iterable of :class:`.Layer` objects.
            defaultLayerName: the name of the default layer of the ones in ``value``.
        """
        if defaultLayerName != DEFAULT_LAYER_NAME:
            import warnings

            warnings.warn(
                "'defaultLayerName' parameter is deprecated; "
                "use Layer.default attribute instead",
                DeprecationWarning,
            )
        layers: dict[str, Layer] = {}
        defaultLayer = None
        for layer in value:
            if not isinstance(layer, Layer):
                raise TypeError(f"expected 'Layer', found '{type(layer).__name__}'")
            if layer.name in layers:
                raise KeyError(f"duplicate layer name: '{layer.name}'")
            if layer.name == defaultLayerName or layer._default:
                if defaultLayer is not None:
                    raise ValueError("more than one layer marked as default")
                if not layer._default:
                    layer._default = True
                defaultLayer = layer
            layers[layer.name] = layer

        if defaultLayer is None:
            raise ValueError("no layer marked as default")
        assert defaultLayer is not None

        return cls(layers=layers, defaultLayer=defaultLayer)

    @classmethod
    def read(cls, reader: UFOReader, lazy: bool = True) -> LayerSet:
        """Instantiates a LayerSet object from a :class:`fontTools.ufoLib.UFOReader`.

        Args:
            path: The path to the UFO to load.
            lazy: If True, load glyphs, data files and images as they are accessed. If
                False, load everything up front.
        """
        layers: dict[str, Layer] = {}
        defaultLayer = None

        defaultLayerName = reader.getDefaultLayerName()

        for layerName in reader.getLayerNames():
            isDefault = layerName == defaultLayerName
            if isDefault or not lazy:
                layer = cls._loadLayer(reader, layerName, lazy, isDefault)
                if isDefault:
                    defaultLayer = layer
                layers[layerName] = layer
            else:
                layers[layerName] = _LAYER_NOT_LOADED

        assert defaultLayer is not None

        self = cls(layers=layers, defaultLayer=defaultLayer)
        self._lazy = lazy
        if lazy:
            self._reader = reader

        return self

    def unlazify(self) -> None:
        """Load all layers into memory."""
        if self._lazy:
            for layer in self:
                layer.unlazify()
        self._lazy = False

    __deepcopy__ = _deepcopy_unlazify_attrs

    __getstate__ = _getstate_unlazify_attrs
    __setstate__ = _setstate_attrs

    @staticmethod
    def _loadLayer(
        reader: UFOReader, layerName: str, lazy: bool = True, default: bool = False
    ) -> Layer:
        glyphSet = reader.getGlyphSet(layerName)
        return Layer.read(layerName, glyphSet, lazy=lazy, default=default)

    def loadLayer(self, layerName: str, lazy: bool = True) -> Layer:
        # XXX: Remove this method and do business via _loadLayer or take this one
        # private.
        assert self._reader is not None
        if layerName not in self._layers:
            raise KeyError(layerName)
        layer = self._loadLayer(self._reader, layerName, lazy)
        self._layers[layerName] = layer
        return layer

    @property
    def defaultLayer(self) -> Layer:
        return self._defaultLayer

    @defaultLayer.setter
    def defaultLayer(self, layer: Layer) -> None:
        if layer is self._defaultLayer:
            return
        if layer not in self._layers.values():
            raise ValueError(
                f"Layer {layer!r} not found in layer set; can't set as default"
            )
        if self._defaultLayer.name == DEFAULT_LAYER_NAME:
            raise ValueError(
                "there's already a layer named 'public.default' which must stay default"
            )
        self._defaultLayer._default = False
        layer._default = True
        self._defaultLayer = layer

    def __contains__(self, name: str) -> bool:
        return name in self._layers

    def __delitem__(self, name: str) -> None:
        if self.defaultLayer is not None:
            if name == self.defaultLayer.name:
                raise KeyError("cannot delete default layer %r" % name)
        del self._layers[name]

    def __getitem__(self, name: str) -> Layer:
        layer_object = self._layers[name]
        if layer_object is _LAYER_NOT_LOADED:
            return self.loadLayer(name)
        return layer_object

    def __iter__(self) -> Iterator[Layer]:
        for layer_name, layer_object in self._layers.items():
            if layer_object is _LAYER_NOT_LOADED:
                yield self.loadLayer(layer_name)
            else:
                yield layer_object

    def __len__(self) -> int:
        return len(self._layers)

    def get(self, name: str, default: T | None = None) -> T | Layer | None:
        try:
            return self[name]
        except KeyError:
            return default

    def keys(self) -> AbstractSet[str]:
        return self._layers.keys()

    def __repr__(self) -> str:
        n = len(self._layers)
        return "<{}.{} ({} layer{}) at {}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            n,
            "s" if n > 1 else "",
            hex(id(self)),
        )

    @property
    def layerOrder(self) -> list[str]:
        """The font's layer order.

        Getter:
            Returns the font's layer order.

        Note:
            The getter always returns a new list, modifications to it do not change
            the LayerSet.

        Setter:
            Sets the font's layer order. The set order value must contain all layers
            that are present in the LayerSet.
        """
        return list(self._layers)

    @layerOrder.setter
    def layerOrder(self, order: list[str]) -> None:
        if set(order) != set(self._layers):
            raise Error(
                "`order` must contain the same layers that are currently present."
            )
        self._layers = {name: self._layers[name] for name in order}

    def newLayer(self, name: str, **kwargs: Any) -> Layer:
        """Creates and returns a named layer.

        Args:
            name: The layer name.
            kwargs: Arguments passed to the constructor of Layer.
        """
        if name in self._layers:
            raise KeyError("layer %r already exists" % name)
        self._layers[name] = layer = Layer(name, **kwargs)
        if layer._default:
            self.defaultLayer = layer
        return layer

    def renameGlyph(self, name: str, newName: str, overwrite: bool = False) -> None:
        """Renames a glyph across all layers.

        Args:
            name: The old name.
            newName: The new name.
            overwrite: If False, raises exception if newName is already taken in any
                layer. If True, overwrites (read: deletes) the old Glyph object.
        """
        # Note: this would be easier if the glyph contained the layers!
        if name == newName:
            return
        # make sure we're copying something
        if not any(name in layer for layer in self):
            raise KeyError("name %r is not in layer set" % name)
        # prepare destination, delete if overwrite=True or error
        for layer in self:
            if newName in layer:
                if overwrite:
                    del layer[newName]
                else:
                    raise KeyError("target name %r already exists" % newName)
        # now do the move
        for layer in self:
            if name in layer:
                layer[newName] = glyph = layer.pop(name)
                glyph._name = newName

    def renameLayer(self, name: str, newName: str, overwrite: bool = False) -> None:
        """Renames a layer.

        Args:
            name: The old name.
            newName: The new name.
            overwrite: If False, raises exception if newName is already taken. If True,
                overwrites (read: deletes) the old Layer object.
        """
        if name == newName:
            return
        if not overwrite and newName in self._layers:
            raise KeyError("target name %r already exists" % newName)
        layer = self[name]
        del self._layers[name]
        self._layers[newName] = layer
        layer._name = newName
        if newName == DEFAULT_LAYER_NAME:
            self.defaultLayer = layer

    def write(self, writer: UFOWriter, saveAs: bool | None = None) -> None:
        """Writes this LayerSet to a :class:`fontTools.ufoLib.UFOWriter`.

        Args:
            writer(fontTools.ufoLib.UFOWriter): The writer to write to.
            saveAs: If True, tells the writer to save out-of-place. If False, tells the
                writer to save in-place. This affects how resources are cleaned before
                writing.
        """
        if saveAs is None:
            saveAs = self._reader is not writer
        # if in-place, remove deleted layers
        layers = self._layers
        if not saveAs:
            for name in set(writer.getLayerNames()).difference(layers):
                writer.deleteGlyphSet(name)
        # write layers
        defaultLayer = self.defaultLayer
        for name, layer in layers.items():
            default = layer is defaultLayer
            if layer is _LAYER_NOT_LOADED:
                if saveAs:
                    layer = self.loadLayer(name, lazy=False)
                else:
                    continue
            glyphSet = writer.getGlyphSet(name, defaultLayer=default)
            layer.write(glyphSet, saveAs=saveAs)
        writer.writeLayerContents(self.layerOrder)

    def _unstructure(self, converter: Converter) -> list[dict[str, Any]]:
        return [converter.unstructure(layer) for layer in self]

    @staticmethod
    def _structure(
        data: list[dict[str, Any]], cls: Type[LayerSet], converter: Converter
    ) -> LayerSet:
        return cls.from_iterable(converter.structure(layer, Layer) for layer in data)



from __future__ import annotations

import sys
from functools import partial
from typing import Any, Callable, Tuple, Type, cast

from attrs import fields, has, resolve_types
from cattrs import Converter
from cattrs.gen import make_dict_structure_fn, make_dict_unstructure_fn, override
from cattrs.gen._consts import AttributeOverride
from fontTools.misc.transform import Transform

is_py37 = sys.version_info[:2] == (3, 7)

if is_py37:

    def get_origin(cl: Type[Any]) -> Any:
        return getattr(cl, "__origin__", None)

else:
    from typing import get_origin  # type: ignore


__all__ = [
    "register_hooks",
    "structure",
    "unstructure",
]


def is_ufoLib2_class(cls: Type[Any]) -> bool:
    mod: str = getattr(cls, "__module__", "")
    return mod.split(".")[0] == "ufoLib2"


def is_ufoLib2_attrs_class(cls: Type[Any]) -> bool:
    return is_ufoLib2_class(cls) and (has(cls) or has(get_origin(cls)))


def is_ufoLib2_class_with_custom_unstructure(cls: Type[Any]) -> bool:
    return is_ufoLib2_class(cls) and hasattr(cls, "_unstructure")


def is_ufoLib2_class_with_custom_structure(cls: Type[Any]) -> bool:
    return is_ufoLib2_class(cls) and hasattr(cls, "_structure")


def register_hooks(conv: Converter, allow_bytes: bool = True) -> None:
    def attrs_hook_factory(
        cls: Type[Any], gen_fn: Callable[..., Callable[..., Any]], structuring: bool
    ) -> Callable[..., Any]:
        base = get_origin(cls)
        if base is None:
            base = cls
        attribs = fields(base)
        # PEP563 postponed annotations need resolving as we check Attribute.type below
        resolve_types(base)
        kwargs: dict[str, bool | AttributeOverride] = {
            "_cattrs_detailed_validation": conv.detailed_validation
        }
        if structuring:
            kwargs["_cattrs_forbid_extra_keys"] = conv.forbid_extra_keys
            kwargs["_cattrs_prefer_attrib_converters"] = conv._prefer_attrib_converters
        else:
            kwargs["_cattrs_omit_if_default"] = conv.omit_if_default
        for a in attribs:
            if a.type in conv.type_overrides:
                # cattrs' gen_(un)structure_attrs_fromdict (used by default for attrs
                # classes that don't have a custom hook registered) check for any
                # type_overrides (Dict[Type, AttributeOverride]); they allow a custom
                # converter to omit specific attributes of given type e.g.:
                # >>> conv = Converter(type_overrides={Image: override(omit=True)})
                attrib_override = conv.type_overrides[a.type]
            else:
                # by default, we omit all Optional attributes (i.e. with None default),
                # overriding a Converter's global 'omit_if_default' option. Specific
                # attibutes can still define their own 'omit_if_default' behavior in
                # the Attribute.metadata dict.
                attrib_override = override(
                    omit_if_default=a.metadata.get(
                        "omit_if_default", a.default is None or None
                    ),
                    rename=a.metadata.get(
                        "rename_attr", a.name[1:] if a.name[0] == "_" else None
                    ),
                    omit=not a.init,
                )
            kwargs[a.name] = attrib_override

        return gen_fn(cls, conv, **kwargs)

    def custom_unstructure_hook_factory(cls: Type[Any]) -> Callable[[Any], Any]:
        return partial(cls._unstructure, converter=conv)

    def custom_structure_hook_factory(cls: Type[Any]) -> Callable[[Any, Any], Any]:
        return partial(cls._structure, converter=conv)

    def unstructure_transform(t: Transform) -> Tuple[float]:
        return cast(Tuple[float], tuple(t))

    def structure_transform(t: Tuple[float], _: Any) -> Transform:
        return Transform(*t)

    conv.register_unstructure_hook_factory(
        is_ufoLib2_attrs_class,
        partial(attrs_hook_factory, gen_fn=make_dict_unstructure_fn, structuring=False),
    )
    conv.register_unstructure_hook_factory(
        is_ufoLib2_class_with_custom_unstructure,
        custom_unstructure_hook_factory,
    )
    conv.register_unstructure_hook(
        cast(Type[Transform], Transform), unstructure_transform
    )

    conv.register_structure_hook(cast(Type[Transform], Transform), structure_transform)
    conv.register_structure_hook_factory(
        is_ufoLib2_attrs_class,
        partial(attrs_hook_factory, gen_fn=make_dict_structure_fn, structuring=True),
    )
    conv.register_structure_hook_factory(
        is_ufoLib2_class_with_custom_structure,
        custom_structure_hook_factory,
    )

    if not allow_bytes:
        from base64 import b64decode, b64encode

        def unstructure_bytes(v: bytes) -> str:
            return (b64encode(v) if v else b"").decode("utf8")

        def structure_bytes(v: str, _: Any) -> bytes:
            return b64decode(v)

        conv.register_unstructure_hook(bytes, unstructure_bytes)
        conv.register_structure_hook(bytes, structure_bytes)


default_converter = Converter(
    omit_if_default=True,
    forbid_extra_keys=True,
    prefer_attrib_converters=False,
)
register_hooks(default_converter, allow_bytes=False)

structure = default_converter.structure
unstructure = default_converter.unstructure

# same as default_converter but allows bytes
binary_converter = Converter(
    omit_if_default=True,
    forbid_extra_keys=True,
    prefer_attrib_converters=False,
)
register_hooks(binary_converter, allow_bytes=True)



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Mapping, Tuple

from ufoLib2.serde import serde

if TYPE_CHECKING:
    from typing import Type

    from cattrs import Converter

KerningPair = Tuple[str, str]


@serde
class Kerning(Dict[KerningPair, float]):
    def as_nested_dicts(self) -> dict[str, dict[str, float]]:
        result: dict[str, dict[str, float]] = {}
        for (left, right), value in self.items():
            result.setdefault(left, {})[right] = value
        return result

    @classmethod
    def from_nested_dicts(self, kerning: Mapping[str, Mapping[str, float]]) -> Kerning:
        return Kerning(
            ((left, right), kerning[left][right])
            for left in kerning
            for right in kerning[left]
        )

    def _unstructure(self, converter: Converter) -> dict[str, dict[str, float]]:
        del converter  # unused
        return self.as_nested_dicts()

    @staticmethod
    def _structure(
        data: Mapping[str, Mapping[str, float]],
        cls: Type[Kerning],
        converter: Converter,
    ) -> Kerning:
        del converter  # unused
        return cls.from_nested_dicts(data)



from __future__ import annotations

import os
from typing import Protocol, TypeVar, Union

from fontTools.pens.basePen import AbstractPen
from fontTools.pens.pointPen import AbstractPointPen

T = TypeVar("T")
"""Generic variable for mypy for trivial generic function signatures."""

PathLike = Union[str, bytes, "os.PathLike[str]", "os.PathLike[bytes]"]
"""Represents a path in various possible forms."""


class Drawable(Protocol):
    """Stand-in for an object that can draw itself with a given pen.

    See :mod:`fontTools.pens.basePen` for an introduction to pens.
    """

    def draw(self, pen: AbstractPen) -> None: ...


class DrawablePoints(Protocol):
    """Stand-in for an object that can draw its points with a given pen.

    See :mod:`fontTools.pens.pointPen` for an introduction to point pens.
    """

    def drawPoints(self, pen: AbstractPointPen) -> None: ...


class HasIdentifier(Protocol):
    """Any object that has a unique identifier in some context that can be
    used as a key in a public.objectLibs dictionary."""

    identifier: str | None


class GlyphSet(Protocol):
    """Any container that holds drawable objects.

    In ufoLib2, this usually refers to :class:`.Font` (referencing glyphs in the
    default layer) and :class:`.Layer` (referencing glyphs in that particular layer).
    Ideally, this would be a simple subclass of ``Mapping[str, Union[Drawable, DrawablePoints]]``,
    but due to historic reasons, the established objects don't conform to ``Mapping``
    exactly.

    The protocol contains what is used in :mod:`fontTools.pens` at v4.18.2
    (grep for ``.glyphSet``).
    """

    # "object" instead of "str" because that's what typeshed says a Mapping should have.
    def __contains__(self, name: object) -> bool: ...

    def __getitem__(self, name: str) -> Drawable | DrawablePoints: ...



from __future__ import annotations

from typing import Any


class Error(Exception):
    """The base exception for ufoLib2."""


class ExtrasNotInstalledError(Error):
    """The extras required for this feature are not installed."""

    def __init__(self, extras: str) -> None:
        super().__init__(f"Extras not installed: ufoLib2[{extras}]")

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        raise self



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from fontTools.misc.transform import Transform
from fontTools.pens.pointPen import AbstractPointPen

from ufoLib2.objects.component import Component
from ufoLib2.objects.contour import Contour
from ufoLib2.objects.point import Point

if TYPE_CHECKING:
    from ufoLib2.objects.glyph import Glyph


class GlyphPointPen(AbstractPointPen):  # type: ignore
    """A point pen.

    See :mod:`fontTools.pens.basePen` and :mod:`fontTools.pens.pointPen` for an
    introduction to pens.
    """

    __slots__ = "_glyph", "_contour"

    def __init__(self, glyph: Glyph) -> None:
        self._glyph: Glyph = glyph
        self._contour: Contour | None = None

    def beginPath(self, identifier: str | None = None, **kwargs: Any) -> None:
        self._contour = Contour(identifier=identifier)

    def endPath(self) -> None:
        if self._contour is None:
            raise ValueError("Call beginPath first.")
        self._glyph.contours.append(self._contour)
        self._contour = None

    def addPoint(
        self,
        pt: tuple[float, float],
        segmentType: str | None = None,
        smooth: bool = False,
        name: str | None = None,
        identifier: str | None = None,
        **kwargs: Any,
    ) -> None:
        if self._contour is None:
            raise ValueError("Call beginPath first.")
        x, y = pt
        self._contour.append(
            Point(
                x, y, type=segmentType, smooth=smooth, name=name, identifier=identifier
            )
        )

    def addComponent(
        self,
        baseGlyph: str,
        transformation: Transform,
        identifier: str | None = None,
        **kwargs: Any,
    ) -> None:
        component = Component(baseGlyph, transformation, identifier=identifier)
        self._glyph.components.append(component)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Mapping, Union, cast

from ufoLib2.constants import DATA_LIB_KEY
from ufoLib2.serde import serde

if TYPE_CHECKING:
    from typing import Type

    from cattrs import Converter

# unfortunately mypy is not smart enough to support recursive types like plist...
# PlistEncodable = Union[
#     bool,
#     bytes,
#     datetime,
#     float,
#     int,
#     str,
#     Mapping[str, PlistEncodable],
#     Sequence[PlistEncodable],
# ]


def _convert_Lib(value: Mapping[str, Any]) -> Lib:
    return value if isinstance(value, Lib) else Lib(value)


# getter/setter properties used by Font, Layer, Glyph
def _get_lib(self: Any) -> Lib:
    return cast(Lib, self._lib)


def _set_lib(self: Any, value: Mapping[str, Any]) -> None:
    self._lib = _convert_Lib(value)


def _get_tempLib(self: Any) -> Lib:
    return cast(Lib, self._tempLib)


def _set_tempLib(self: Any, value: Mapping[str, Any]) -> None:
    self._tempLib = _convert_Lib(value)


def is_data_dict(value: Any) -> bool:
    return (
        isinstance(value, Mapping)
        and "type" in value
        and value["type"] == DATA_LIB_KEY
        and "data" in value
    )


def _unstructure_data(value: Any, converter: Converter) -> Any:
    if isinstance(value, bytes):
        return {"type": DATA_LIB_KEY, "data": converter.unstructure(value)}
    elif isinstance(value, (list, tuple)):
        return [_unstructure_data(v, converter) for v in value]
    elif isinstance(value, Mapping):
        return {k: _unstructure_data(v, converter) for k, v in value.items()}
    return value


def _structure_data_inplace(
    key: Union[int, str], value: Any, container: Any, converter: Converter
) -> None:
    if isinstance(value, list):
        for i, v in enumerate(value):
            _structure_data_inplace(i, v, value, converter)
    elif is_data_dict(value):
        container[key] = converter.structure(value["data"], bytes)
    elif isinstance(value, Mapping):
        for k, v in value.items():
            _structure_data_inplace(k, v, value, converter)


@serde
class Lib(Dict[str, Any]):
    def _unstructure(self, converter: Converter) -> dict[str, Any]:
        # avoid encoding if converter supports bytes natively
        test = converter.unstructure(b"\0")
        if isinstance(test, bytes):
            return dict(self)
        elif not isinstance(test, str):
            raise NotImplementedError(type(test))

        data: dict[str, Any] = _unstructure_data(self, converter)
        return data

    @staticmethod
    def _structure(
        data: Mapping[str, Any],
        cls: Type[Lib],
        converter: Converter,
    ) -> Lib:
        self = cls(data)
        for k, v in self.items():
            _structure_data_inplace(k, v, self, converter)
        return self



from __future__ import annotations

from ufoLib2.objects.anchor import Anchor
from ufoLib2.objects.component import Component
from ufoLib2.objects.contour import Contour
from ufoLib2.objects.dataSet import DataSet
from ufoLib2.objects.features import Features
from ufoLib2.objects.font import Font
from ufoLib2.objects.glyph import Glyph
from ufoLib2.objects.guideline import Guideline
from ufoLib2.objects.image import Image
from ufoLib2.objects.imageSet import ImageSet
from ufoLib2.objects.info import Info
from ufoLib2.objects.kerning import Kerning
from ufoLib2.objects.layer import Layer
from ufoLib2.objects.layerSet import LayerSet
from ufoLib2.objects.lib import Lib
from ufoLib2.objects.point import Point

__all__ = [
    "Anchor",
    "Component",
    "Contour",
    "DataSet",
    "Features",
    "Font",
    "Glyph",
    "Guideline",
    "Image",
    "ImageSet",
    "Info",
    "Kerning",
    "Layer",
    "LayerSet",
    "Lib",
    "Point",
]



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Type

from attrs import define

from ufoLib2.serde import serde

if TYPE_CHECKING:
    from cattrs import Converter


RE_NEWLINES = re.compile(r"\r\n|\r")


@serde
@define
class Features:
    """A data class representing UFO features.

    See http://unifiedfontobject.org/versions/ufo3/features.fea/.
    """

    text: str = ""
    """Holds the content of the features.fea file."""

    def __bool__(self) -> bool:
        return bool(self.text)

    def __str__(self) -> str:
        return self.text

    def normalize_newlines(self) -> Features:
        """Normalize CRLF and CR newlines to just LF."""
        self.text = RE_NEWLINES.sub("\n", self.text)
        return self

    def _unstructure(self, converter: Converter) -> str:
        del converter  # unused
        return self.text

    @staticmethod
    def _structure(data: str, cls: Type[Features], converter: Converter) -> Features:
        del converter  # unused
        return cls(data)



from __future__ import annotations

from typing import Optional

from attrs import define

from ufoLib2.objects.misc import AttrDictMixin
from ufoLib2.serde import serde


@serde
@define
class Guideline(AttrDictMixin):
    """Represents a single guideline.

    See http://unifiedfontobject.org/versions/ufo3/glyphs/glif/#guideline. Has some
    data composition restrictions.
    """

    x: Optional[float] = None
    """The origin x coordinate of the guideline."""

    y: Optional[float] = None
    """The origin y coordinate of the guideline."""

    angle: Optional[float] = None
    """The angle of the guideline."""

    name: Optional[str] = None
    """The name of the guideline, no uniqueness required."""

    color: Optional[str] = None
    """The color of the guideline."""

    identifier: Optional[str] = None
    """The globally unique identifier of the guideline."""

    def __attrs_post_init__(self) -> None:
        x, y, angle = self.x, self.y, self.angle
        if x is None and y is None:
            raise ValueError("x or y must be present")
        if x is None or y is None:
            if angle is not None:
                raise ValueError("if 'x' or 'y' are None, 'angle' must not be present")
        if x is not None and y is not None and angle is None:
            raise ValueError("if 'x' and 'y' are defined, 'angle' must be defined")
        if angle is not None and not (0 <= angle <= 360):
            raise ValueError("angle must be between 0 and 360")
